@@ -1,34 +1,32 @@
-/*\
-||| Written By Fredrik Hbinette <hubbe@hubbe.net>
-||| and Louis Bavoil <louis@bavoil.net>
-||| No warrenties, use at your own risk.
-\*/
-
-#include <unistd.h>
-#include <string.h>
-#include <stdio.h>
-#include <sysexits.h>
-#include <signal.h>
-#include <stdarg.h>
-#include <npapi.h>
-#include <sys/stat.h>
-#include <sys/wait.h>
-#include <errno.h>
-#include <sys/time.h>
-#include <X11/X.h>
-#include <X11/Xos.h>
-#include <X11/Intrinsic.h>
-#include <X11/Xatom.h>
-#include <X11/Xlib.h>
-#include <sys/socket.h>
+/*****************************************************************************
+ * Authors: Louis Bavoil <louis@bavoil.net>
+ *          Fredrik Hbinette <hubbe@hubbe.net>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
+*****************************************************************************/
 
 #include "mozplugger.h"
 
 #define WINDOW ((Window) windata.window)
-#define FEND(X,Y) (strchr(X,Y)?strchr(X,Y):(X)+strlen(X))
 
-Display *display = 0;
-char *displayname = 0;
+#ifndef MIN
+#define MIN(X,Y) ((X)<(Y)?(X):(Y))
+#endif
+
+Display *display;
+char *displayname;
 NPWindow windata;
 int flags;
 int repeats;
@@ -37,22 +35,28 @@ char *command;
 char *file;
 char *mimetype;
 
-static char *my_strstr(char *s1,const char *s2)
+static XWindowAttributes wattr;
+static Window victim;
+static Window old_parent;
+
+#define MAX_IGNORED 1000
+static int windows_found;
+static Window ignored[MAX_IGNORED];
+
+static int got_sigwinch;
+
+/*****************************************************************************
+ * Functions below are called by the child process.
+ *****************************************************************************/
+void sigwinch(int tmp)
 {
-     for (;*s1;s1++)
-     {
-	  const char *p1,*p2;
-	  for (p1 = s1, p2 = s2; *p2; p1++, p2++)
-	       if (*p2 != *p1) break;
-	  if (!*p2) return s1;
-     }
-     return NULL;
+     D("GOT SIGWINCH\n");
+     got_sigwinch = 1;
 }
 
-/*
- * Functions below are called by the child process
- */
-
+/*****************************************************************************
+ * Called on X11 errors.
+ *****************************************************************************/
 int error_handler(Display *dpy, XErrorEvent *err)
 {
      char s[1024];
@@ -61,31 +65,28 @@ int error_handler(Display *dpy, XErrorEvent *err)
      return 0;
 }
 
-static XWindowAttributes wattr;
-static int xaspect;
-static int yaspect;
-static Window victim = 0;
-static Window old_parent = 0;
-static Atom lock_property;
-
-static void place_window(void)
+/*****************************************************************************
+ * Make the window swallowed by the browser.
+ *****************************************************************************/
+static void place_window()
 {
      int x;
      Window parent = WINDOW;
      int w = windata.width;
      int h = windata.height;
-     D("Old: %x  New:%x\n",old_parent,parent);
+
+     D("Old: %x  New:%x\n", old_parent, parent);
+
      if (old_parent != parent)
      {
-	  D("Withdrawing Window %x\n",victim);
-	  XWithdrawWindow(display, victim,
-			  XScreenNumberOfScreen (wattr.screen));
+	  D("Withdrawing Window %x\n", victim);
+	  XWithdrawWindow(display, victim, XScreenNumberOfScreen (wattr.screen));
 	  XSync(display, FALSE);
 
 	  if (!old_parent)
 	  {
 	       XWMHints *leader_change;
-	       D("Changing leader of window %x\n",victim);
+	       D("Changing leader of window %x\n", victim);
 	       if ((leader_change = XGetWMHints(display, victim)))
 	       {
 		    leader_change->flags = (leader_change->flags | WindowGroupHint);
@@ -96,195 +97,25 @@ static void place_window(void)
 	       }
 	  }
 
-	  for (x=0; x<25; x++)
+	  D("Reparenting window %x into %x\n", victim, parent);
+	  for (x = 0; x < 25; x++)
 	  {
-	       D("Reparenting window %x into %x\n",victim,WINDOW);
-	       XReparentWindow(display, victim, WINDOW, 0, 0);
+	       XReparentWindow(display, victim, parent, 0, 0);
 	       XSync(display, FALSE);
 	  }
      }
 
-     if (flags & H_FILL)
-     {
-	  D("Resizing window %x with FILL\n",victim);
-	  D("New size: %dx%d+%d+%d (of %dx%d)\n",w,h,
-	    0,0,
-	    w,h);
-    
-	  XMoveResizeWindow(display, victim,
-			    0,0,
-			    w,h);
-     }
-     else if (flags & H_MAXASPECT)
-     {
-	  int tmpw, tmph;
-	  D("Resizing window %x with MAXASPECT\n",victim);
-	  if (xaspect && yaspect)
-	  {
-	       tmph = h/yaspect;
-	       tmpw = w/xaspect;
-	       if (tmpw<tmph) tmph = tmpw;
-	       tmpw = tmph*xaspect;
-	       tmph = tmph*yaspect;
-      
-	       D("New size: %dx%d+%d+%d (of %dx%d)\n",tmpw,tmph,
-		 (w - tmpw)/2,
-		 (h - tmph)/2,
-		 w,h);
-	       XMoveResizeWindow(display, victim,
-				 (w - tmpw)/2,
-				 (h - tmph)/2,
-				 tmpw, 
-				 tmph);
-	  }
-     }
-
      XMapWindow(display, victim);
      XSync(display, FALSE);
 
+     maximize_window(flags, display, victim, w, h);
+
      D("placewindow done\n");
 }
 
-/* X11 locking mechanism */
-static void my_xlockwindow(Display *dpy,
-			   Window win,
-			   Atom prop)
-{
-     char buffer[1024];
-     struct timeval tv, tv2;
-
-     gettimeofday(&tv, 0);
-     gethostname(buffer,sizeof(buffer)-1);
-     sprintf(buffer+strlen(buffer), "-%d-%ld:%ld!",
-	     getpid(), (long)tv.tv_sec, (long)tv.tv_usec);
-
-     tv.tv_sec += 8; /* Max lock wait time, 8 seconds */
-     D("Attempting to lock property = %d!\n", prop);
-     D("My LOCKID='%s'\n", buffer);
-     while(1)
-     {
-	  int mode = PropModeAppend;
-	  int fmt;
-	  unsigned long nitems = 0, bytes = 0;
-	  unsigned char *property = 0;
-	  Atom type;
-
-	  while(1)
-	  {
-	       property = 0;
-	       XGetWindowProperty(dpy, win, prop,
-				  0, 1,
-				  0, /* Delete */
-				  XA_STRING,
-				  &type, &fmt, &nitems, &bytes,
-				  &property);
-	       if (!property) break;
-	       D("Property already set, bytes_left=%d, waiting..\n",bytes);
-	       XFree(property);
-	       gettimeofday(&tv2, 0);
-	       if (tv2.tv_sec > tv.tv_sec)
-	       {
-		    D("Stealing lock....\n");
-		    /* Time to steal the lock! */
-		    mode = PropModeReplace;
-		    gettimeofday(&tv, 0);
-		    INC_TIME(&tv, 0, 10000 + (rand() & 16383));
-		    break;
-	       }
-	       usleep((rand() & 16383) + 2000); /* 0.18383 seconds */
-	  }
-    
-	  D("XChangeProperty, mode=%d\n",mode);
-	  XChangeProperty(dpy, win, prop, XA_STRING, 8,
-			  mode, buffer, strlen(buffer));
-
-	  D("Getting property, did we get the lock?\n");
-	  property = 0;
-	  XGetWindowProperty(dpy, win, prop,
-			     0, sizeof(buffer),
-			     0, /* Delete */
-			     XA_STRING,
-			     &type, &fmt, &nitems, &bytes,
-			     &property);
-	  if (property)
-	  {
-	       int done =! strncmp(buffer,property,strlen(buffer));
-	       D("LOCK: done=%d, bytes_left=%d\n",done,bytes);
-	       XFree(property);
-	       if (done) return;
-	  }
-     }
-}
-
-static void my_xunlockwindow(Display *dpy,
-			     Window win,
-			     Atom prop)
-{
-     D("!!!UNLOCKING!!!\n");
-     XDeleteProperty(dpy, win, prop);
-}
-
-static int got_sigwinch = 0;
-
-static void sigwinch(int tmp)
-{
-     D("GOT SIGWINCH\n");
-     got_sigwinch = 1;
-}
-
-static void run_app(char **argv)
-{
-#ifdef H_NOISY
-     int nl;
-     if (flags & H_NOISY)
-     {
-	  D("Redirecting stdout and stderr\n");
-	  if ((nl = open("/dev/null", O_RDONLY)) == -1)
-	       exit(EX_UNAVAILABLE);
-	  dup2(nl,1);
-	  dup2(nl,2);
-	  close(nl);
-     }
-#endif
-     
-     execvp(argv[0],argv);
-     D("Execvp failed..%d\n",errno);
-     exit(EX_UNAVAILABLE);
-}
-
-static int gcd(int a, int b)
-{
-     if (a < b) return gcd(b,a);
-     if (b == 0) return a;
-     return gcd(b, a % b);
-}
-
-static void max_aspect()
-{
-     int tmp;
-     XWindowAttributes ca;
-     D("Max aspect\n");
-     XGetWindowAttributes(display, victim, &ca);
-     while (1)
-     {
-	  tmp = gcd(ca.width, ca.height);
-	  xaspect = ca.width/tmp;
-	  yaspect = ca.height/tmp;
-	  if (xaspect>1 && yaspect>1 && (xaspect+yaspect>100))
-	  {
-	       xaspect >>= 1;
-	       xaspect >>= 1;
-	  } else {
-	       break;
-	  }
-	  D("xaspect=%d yaspect=%d\n", xaspect, yaspect);
-     }
-}
-
-#define MAX_IGNORED 1000
-static int windows_found = 0;
-static Window ignored[MAX_IGNORED];
-
+/*****************************************************************************
+ * Helper functions for winrecur().
+ *****************************************************************************/
 static int inital_win_cb(Window w, int depth)
 {
      if (windows_found < MAX_IGNORED)
@@ -295,13 +126,13 @@ static int inital_win_cb(Window w, int depth)
 static int win_cb(Window w, int depth)
 {
      int q;
-     for (q=0;q<windows_found;q++)
+     for (q = 0; q < windows_found; q++)
 	  if (ignored[q] == w)
 	       return 0;
      return 1;
 }
 
-static int is_viewable(Window w)
+static int is_viewable(Display *display, Window w)
 {
      XWindowAttributes a;
 
@@ -318,6 +149,9 @@ static int is_viewable(Window w)
      return 1;
 }
 
+/*****************************************************************************
+ * Find the window to swallow by scanning a window tree recursively.
+ *****************************************************************************/
 static Window winrecur(Window from,
 		       int depth,
 		       char *name,
@@ -325,7 +159,7 @@ static Window winrecur(Window from,
 {
      Window root, parent;
      Window *children = NULL;
-     int e,num_children;
+     int e, num_children;
      Window found = -1;
   
      D("Winrecur, checking window %x (depth=%d)\n", from, depth);
@@ -338,7 +172,7 @@ static Window winrecur(Window from,
 
      D("Num children = %d\n", num_children);
 
-     for (e=0; e<num_children; e++)
+     for (e = 0; e < num_children; e++)
      {
 	  char *windowname;
 	  unsigned long nitems=0, bytes=0; 
@@ -346,7 +180,7 @@ static Window winrecur(Window from,
 	  int fmt;
 	  Atom type;
 
-	  if (!is_viewable(children[e]))
+	  if (!is_viewable(display, children[e]))
 	       continue;
 
 	  D("XFetchName\n");
@@ -362,7 +196,7 @@ static Window winrecur(Window from,
 	  }
 
 	  /* Return the last window found */
-	  if (found != -1) return found;
+	  if (found != -1) goto success;
 
 	  D("XGetTextProperty\n");
 	  property = 0;
@@ -381,7 +215,7 @@ static Window winrecur(Window from,
 		    char *n = (char *)property;
 		    while (len)
 		    {
-			 D("Winrecur, checking window CLASS %x (%s != %s)\n",children[e],n,name);
+			 D("Winrecur, checking window CLASS %x (%s != %s)\n", children[e], n,name);
 			 if (!name || !strcmp(n, name))
 			      if (cb(children[e],depth))
 				   found = children[e];
@@ -393,78 +227,179 @@ static Window winrecur(Window from,
 	       XFree(property);
 	  }
      }
-     if (found != -1) return found;
-     if (depth>0)
+     if (found != -1) goto success;
+     if (depth > 0)
      {
-	  for (e=0; e<num_children; e++)
+	  for (e = 0;  e < num_children; e++)
 	  {
 	       Window ret = winrecur(children[e], depth-1, name, cb);
 	       if (ret) found = ret;
 	  }
      }
-     if (found != -1) return found;
+     if (found != -1) goto success;
      XFree(children);
      return (Window)0;
+
+ success:
+     XFree(children);
+     return found;
 }
 
+/*****************************************************************************
+ * Swallow using winrecur()
+ *****************************************************************************/
 static int rec_swallow()
 {
      if ((victim = winrecur(wattr.root, 3, winname, win_cb)))
      {
-	  my_xunlockwindow(display, wattr.root, lock_property);
-	      
-	  D("Found it!! Victim=%x\n",victim);
-	  if (flags & H_MAXASPECT)
-	       max_aspect();
-	  
+	  D("Found it!! Victim=%x\n", victim);
 	  place_window();
 	  return 1;
      }
      return 0;
 }
 
-static void setup_swallow()
+/*****************************************************************************
+ * Swallow using XCheckMaskEvent()
+ *****************************************************************************/
+static int event_swallow()
 {
-     D("Setting up X data (%s)\n",displayname);
+     XEvent evt;
+     XClassHint windowclass;
+     char *windowname;
+     int found = 0;
+     int len;
+     char *victimname = winname;
+
+     D("Looking for victim... (%s)\n", victimname);
+     /* XCheckMaskEvent is a non-blocking call, go through all
+	waiting SubStructureEvents, untill we're out of them. */
+     while (XCheckMaskEvent(display,SubstructureNotifyMask,&evt))
+     {
+	  /* got a structure event */
+	  D("Got Structure Event of type %x\n", evt.type);
+	  if (evt.type == CreateNotify)
+	  {
+	       /* looks promising, let's see if it's the one we're looking for */
+	       D("Found CreateNotify Event %x\n", evt.type);
+
+	       if (!XGetClassHint(display, evt.xcreatewindow.window, &windowclass))
+	       {
+		    D("Could Not get Class Hint, trying the next one");
+		    continue;
+	       }
+
+	       if (!XFetchName(display, evt.xcreatewindow.window, &windowname))
+	       {
+		    D("Could Not Get Window Name, trying the next one");
+		    continue;
+	       }
+
+	       if (windowname)
+	       {
+		    D("Victim Name is :%s\n", victimname);
+		    D("Window Name is :%s\n", windowname);
+		    len = strlen(victimname);
+		    if (strncmp(victimname, windowname, len) == 0 ||
+			strncmp(victimname, windowclass.res_class, len) == 0 ||
+			strncmp(victimname, windowclass.res_name, len) == 0)
+		    {
+			 victim = evt.xcreatewindow.window;
+			 found = 1;
+			 D("Found it!! Victim=%x\n", victim);
+			 D("Flags %x\n", flags);
+			 place_window();
+		    }
+		    XFree(windowname);
+	       }
+	       if (found) break;
+	  }
+     }
+
+     return found;
+}
+
+
+/*****************************************************************************
+ * Prepare the swallowing.
+ *****************************************************************************/
+static int init_swallow()
+{
+     D("Setting up X data\n");
 
      if (!display)
-     {
-	  display = XOpenDisplay(displayname);
-	  XSetErrorHandler(error_handler);
-     }
-     D("display=%x\n",display);
+	  return 1;
 
-     if (display && XGetWindowAttributes(display, WINDOW, &wattr))
+     D("display=%x\n", display);
+
+     if (!XGetWindowAttributes(display, WINDOW, &wattr))
+          return 1;
+
+     D("rootwin=%x\n", wattr.root);
+
+     victim = 0;
+     old_parent = 0;
+
+     if (flags & H_EVENT_SWALLOW)
      {
-	  victim = 0;
-	  old_parent = 0;
-	  xaspect = yaspect = 0;
-	  D("rootwin=%x\n", wattr.root);
-	  lock_property = XInternAtom(display, "PLUGGER_SWALLOW_LOCK", 0);
-	  D("lock_property=%x\n", lock_property);
-	  my_xlockwindow(display, wattr.root, lock_property);
-	  windows_found = 0;
-	  winrecur(wattr.root, 3, winname, inital_win_cb);
-	  D("Done setting up X data\n");
+          if (!XSelectInput(display, DefaultRootWindow(display), SubstructureNotifyMask))
+	       return 1;
      }
      else
      {
-	  D("setup_swallow() failed!!!!\n");
-	  flags &=~ H_SWALLOW;
+	  windows_found = 0;
+	  winrecur(wattr.root, 3, winname, inital_win_cb);
      }
+
+     D("Done setting up X data\n");
+
+     return 0;
 }
 
+/*****************************************************************************
+ * Terminate the swallowing.
+ *****************************************************************************/
+static void end_swallow()
+{
+     if (flags & H_EVENT_SWALLOW)
+	  XSelectInput(display, DefaultRootWindow(display), 0);
+}
+
+/*****************************************************************************
+ * Wrapper for execvp().
+ *****************************************************************************/
+static void run_app(char **argv)
+{
+     int nl;
+
+     if (flags & H_NOISY)
+     {
+	  D("Redirecting stdout and stderr\n");
+	  if ((nl = open("/dev/null", O_RDONLY)) == -1)
+	       exit(EX_UNAVAILABLE);
+	  dup2(nl, 1);
+	  dup2(nl, 2);
+	  close(nl);
+     }
+     
+     execvp(argv[0],argv);
+     D("Execvp failed..%d\n", errno);
+     exit(EX_UNAVAILABLE);
+}
+
+/*****************************************************************************
+ * Handle the swallowing, the SIGWINCH signal, and wait.
+ *****************************************************************************/
 static void handle_app(int pid)
 {
      int waitpid_done = 0;
-     int unlock_done = 0;
+     int swallow_done = 0;
      int status;
+     int tmp;
 
-     if (flags & H_SWALLOW)
+     if ((flags & H_SWALLOW) || (flags & H_EVENT_SWALLOW))
      {
-	  int tmp;
-	  
-	  for (tmp=0; tmp<2000; tmp++)
+	  for (tmp = 0; tmp < 2000; tmp++)
 	  {
 	       if (!(flags & H_EXITS))
 	       {
@@ -475,44 +410,46 @@ static void handle_app(int pid)
 		    } 
 	       }
 
-	       unlock_done = rec_swallow();
-	       if (unlock_done) break;
+	       if (flags & H_SWALLOW)
+		    swallow_done = rec_swallow();
+	       else
+		    swallow_done = event_swallow();
+
+	       if (swallow_done) break;
 	       
 	       /* Do not increase this wait time or the swallow will
 		  not work correctly on some computers. */
 	       usleep(1000);
 	  }
-	  if (!unlock_done)
-	  {
-	       my_xunlockwindow(display, wattr.root, lock_property);
-	       unlock_done = 1;
-	  }
+	  end_swallow();
      }
      
      while (!waitpid_done)
      {
 	  if (got_sigwinch)
 	  {
-	       Window oldwindow=WINDOW;
+	       Window oldwindow = WINDOW;
 	       got_sigwinch = 0;
 	  
-	       D("Got SIGWINCH, new parent=%x\n",WINDOW);
+	       D("Got SIGWINCH, new parent=%x\n", WINDOW);
 	  
-	       if (victim)
+	       if (swallow_done)
 	       {
 		    place_window();
 	       }
-	       else if (my_strstr(command,"$window") && oldwindow != WINDOW)
+	       else if (strstr(command,"$window") && oldwindow != WINDOW)
 	       {
-		    kill(pid, SIGKILL);
+		    my_kill(pid);
 		    waitpid_done++;
 		    continue;
 	       }
 	  }
-	  D("Waiting for godot... %d\n",pid);
+
 	  if (flags & H_EXITS)
 	       exit(0);
-	  else if (waitpid(pid,&status,0) > 0)
+
+	  D("Waiting for godot... %d\n",pid);
+	  if (waitpid(pid, &status, 0) > 0)
 	       waitpid_done++;
 	  D("Wait done?? (%d)\n",waitpid_done);
      }
@@ -524,7 +461,7 @@ static void handle_app(int pid)
      }
      if (WEXITSTATUS(status) && !(flags & H_IGNORE_ERRORS))
      {
-	  D("Process exited with error code: %d\n",WEXITSTATUS(status));
+	  D("Process exited with error code: %d\n", WEXITSTATUS(status));
 	  exit(WEXITSTATUS(status));
      }
      D("exited ok!\n");
@@ -583,10 +520,9 @@ int main(int argc, char **argv)
 	       loops = repeats;
 	  }
 
-	  /* This application will play the data forever */
+	  /* Expecting the application to loop */
 	  if (flags & H_LOOP)
 	  {
-	       D("Expecting application to loop.\n");
 	       loops = 0x7fffffff;
 	  }
 
@@ -594,7 +530,7 @@ int main(int argc, char **argv)
 	  if ((flags & H_MANY) && !(flags & H_REPEATCOUNT))
 	  {
 	       int e;
-	       sprintf(foo,"file=%s",file);
+	       snprintf(foo, sizeof(buffer), "file=%s", file);
 	       loops = MIN(repeats,10);
 	       for (e=0; e<loops; e++)
 	       {
@@ -602,35 +538,35 @@ int main(int argc, char **argv)
 		    strcat(foo,file);
 	       }
 	  } else {
-	       sprintf(foo,"file=%s",file);
+	       snprintf(foo, sizeof(buffer), "file=%s", file);
 	  }
 	  putenv(foo);
 	  foo+=strlen(foo)+1;
 
 	  /* setup environment variable $env */
-	  sprintf(foo,"window=%ld",(long)WINDOW);
+	  snprintf(foo, sizeof(buffer), "window=%ld", (long)WINDOW);
 	  putenv(foo);
 	  foo+=strlen(foo)+1;
 
 	  /* setup environment variable $repeat */
-	  sprintf(foo,"repeat=%ld",(long)repeats);
+	  snprintf(foo, sizeof(buffer), "repeat=%ld", (long)repeats);
 	  putenv(foo);
 	  foo+=strlen(foo)+1;
 
 	  /* setup environment variable $width */
-	  sprintf(foo,"width=%ld",(long)windata.width);
+	  snprintf(foo, sizeof(buffer), "width=%ld", (long)windata.width);
 	  putenv(foo);
 	  foo+=strlen(foo)+1;
 
 	  /* setup environment variable $height */
-	  sprintf(foo,"height=%ld",(long)windata.height);
+	  snprintf(foo, sizeof(buffer), "height=%ld", (long)windata.height);
 	  putenv(foo);
 	  foo+=strlen(foo)+1;
 
 	  /* setup environment variable $DISPLAY */
 	  if (displayname)
 	  {
-	       sprintf(foo,"DISPLAY=%s",displayname);
+	       snprintf(foo, sizeof(buffer), "DISPLAY=%s", displayname);
 	       D("putenv(%s)\n",foo);
 	       putenv(foo);
 	       foo+=strlen(foo)+1;
@@ -639,24 +575,35 @@ int main(int argc, char **argv)
 	  /* setup environment variable $mimetype */
 	  if (mimetype)
 	  {
-	       sprintf(foo,"mimetype=%s",mimetype);
+	       snprintf(foo, sizeof(buffer), "mimetype=%s", mimetype);
 	       D("putenv(%s)\n",foo);
 	       putenv(foo);
 	       foo+=strlen(foo)+1;
 	  }
 
 	  /* Create command line */
-	  argv[0]="/bin/sh";
-	  argv[1]="-c";
-	  argv[2]=command;
-	  argv[3]=0;
+	  argv[0] = "/bin/sh";
+	  argv[1] = "-c";
+	  argv[2] = command;
+	  argv[3] = NULL;
 
 	  D("Execing %s (repeats=%d loops=%d)\n",command,repeats,loops);
 
-	  if (flags & H_SWALLOW)
-	       setup_swallow();
+	  if ((flags & H_SWALLOW) || (flags & H_EVENT_SWALLOW))
+	  {
+	       XSetErrorHandler(error_handler);
+
+	       if (!display) display = XOpenDisplay(displayname);
+
+	       if (init_swallow())
+	       {
+		    D("setup_swallow() failed!!!!\n");
+		    flags &=~ H_SWALLOW;
+		    flags &=~ H_EVENT_SWALLOW;
+	       }
+	  }
     
-	  D("Running %s\n",command);
+	  D("Running %s\n", command);
 
 	  if ((pid = fork()) == -1)
 	       exit(10);
